#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <string.h>
#include "nrf_log.h"

#include "MegaSig_BLE_Function.h"
#include "MegaSig_Global.h"
#include "MegaSig_BLE_Gattc.h"

/** Event functions */

/**@brief Function called on BLE_GAP_EVT_CONNECTED event.
 *
 * @details Update connection state and proceed to discovering the peer's GATT services.
 *
 * @param[in] p_ble_gap_evt GAP event.
 */
void on_connected(const ble_gap_evt_t *const p_ble_gap_evt)
{
  usb_printf_variadic("Connection established\n");

  m_connected_devices++;
  m_connection_handle = p_ble_gap_evt->conn_handle;
  NRF_LOG_INFO("Connected to device with handle: 0x%04X\n", m_connection_handle);
  m_connection_is_in_progress = false;

  ble_usb_add_string("Connect Device Address: ", 25);
  for (uint16_t i = 0; i < 6; i++)
  {
    m_connected_addr[i] = p_ble_gap_evt->params.connected.peer_addr.addr[i];
  }
  usb_printf_variadic("Connect Device Address: %s\n", m_connected_addr);
}

/**@brief Function called on BLE_GAP_EVT_ADV_REPORT event.
 *
 * @details Create a connection if received advertising packet corresponds to desired BLE device.
 *
 * @param[in] p_ble_gap_evt Advertising Report Event.
 */
void on_adv_report(const ble_gap_evt_t *const p_ble_gap_evt)
{
  uint32_t err_code = 0;
  uint8_t str[STRING_BUFFER_SIZE] = {0};

  // Log the Bluetooth device address of advertisement packet received.
  ble_address_to_string_convert(p_ble_gap_evt->params.adv_report.peer_addr, str);
  current_scan_info = p_ble_gap_evt->params.adv_report;
  uint8_t *address = current_scan_info.peer_addr.addr;
  usb_printf_variadic("Advertising Report: Device Address: %02x:%02x:%02x:%02x:%02x:%02x\n", address[5], address[4], address[3], address[2], address[1], address[0]);
  usb_printf_variadic("Addr Type: 0x%d\n", p_ble_gap_evt->params.adv_report.peer_addr.addr_type);
  usb_printf_variadic("Addr ID Peer: 0x%d\n", p_ble_gap_evt->params.adv_report.peer_addr.addr_id_peer);
  usb_printf_variadic("Device Rssi: %d\n", p_ble_gap_evt->params.adv_report.rssi);
  // usb_printf_variadic("Advertising Data: %s", p_ble_gap_evt->params.adv_report.data.p_data);

  err_code = sd_ble_gap_scan_start(NULL, &m_adv_report_data);

  if (err_code != NRF_SUCCESS)
  {
    usb_printf_variadic("Error starting scan. Error code 0x%02X\n", err_code);
  }
  else
  {
    usb_printf_variadic("Scan starter\n");
  }
}

/**@brief Function called on BLE_GAP_EVT_TIMEOUT event.
 *
 * @param[in] ble_gap_evt_t Timeout Event.
 */
void on_timeout(const ble_gap_evt_t *const p_ble_gap_evt)
{
  if (p_ble_gap_evt->params.timeout.src == BLE_GAP_TIMEOUT_SRC_CONN)
  {
    m_connection_is_in_progress = false;
  }
  else if (p_ble_gap_evt->params.timeout.src == BLE_GAP_TIMEOUT_SRC_SCAN)
  {
    scan_start();
  }
}

/**@brief Function called on BLE_GATTC_EVT_PRIM_SRVC_DISC_RSP event.
 *
 * @details Update service state and proceed to discovering the service's GATT characteristics.
 *
 * @param[in] p_ble_gattc_evt Primary Service Discovery Response Event.
 */
void on_service_discovery_response(const ble_gattc_evt_t *const p_ble_gattc_evt)
{
  int count;
  int service_index;

  count = p_ble_gattc_evt->params.prim_srvc_disc_rsp.count;

  if (p_ble_gattc_evt->gatt_status != NRF_SUCCESS)
  {
    if (p_ble_gattc_evt->gatt_status == 0x010A)
    {
      // service_discovery_start(0x01);
    }
    else
    {
      BLE_USB_PRINT_MSG_NUM("Service discovery failed. Error code 0x", p_ble_gattc_evt->gatt_status, 1);
      return;
    }
  }

  if (count == 0)
  {
    BLE_USB_PRINT_MSG("Service not found\n");
    return;
  }

  if (count > 0)
  {
    ble_gattc_service_t temp;
    for (service_index = 0; service_index < count; service_index++)
    {
      ble_usb_add_string("Service Info: Service UUID = ", 30);
      temp = p_ble_gattc_evt->params.prim_srvc_disc_rsp.services[service_index];
      // //printf("Discovered NUS service. UUID: 0x%04X, "
      //        "start handle: 0x%04X, end handle: 0x%04X, uuid type:0x%04X\n",
      //        temp.uuid.uuid, temp.handle_range.start_handle, temp.handle_range.end_handle, temp.uuid.type);
      ble_usb_add_value_uint16(temp.uuid.uuid, 1);
      ble_usb_add_string(", Service UUID Type = ", 23);
      ble_usb_add_value_uint8(temp.uuid.type, 1);
      ble_usb_add_string(", Service Start Handle = ", 26);
      ble_usb_add_value_uint16(temp.handle_range.start_handle, 1);
      ble_usb_add_string(", Service End Handle = ", 24);
      ble_usb_add_value_uint16(temp.handle_range.end_handle, 1);
      ble_usb_add_string("\n", 1);
      ble_usb_printf();
    }
  }
}

/**@brief Function called on BLE_GATTC_EVT_CHAR_DISC_RSP event.
 *
 * @details Update characteristic state and proceed to discovering the characteristicss descriptors.
 *
 * @param[in] p_ble_gattc_evt Characteristic Discovery Response Event.
 */
void on_characteristic_discovery_response(const ble_gattc_evt_t *const p_ble_gattc_evt)
{
  int count = p_ble_gattc_evt->params.char_disc_rsp.count;
  uint32_t err_code = NRF_SUCCESS;

  if (p_ble_gattc_evt->gatt_status == BLE_GATT_STATUS_ATTERR_ATTRIBUTE_NOT_FOUND)
  {
    // printf("Characteristic discovery complete\n");
    //  sd_ble_gattc_descriptors_discover(m_connection_handle,m_start_handle);
    return;
  }
  else if (p_ble_gattc_evt->gatt_status != NRF_SUCCESS)
  {
    // printf("Characteristic discovery failed. Error code 0x%04X\n",
    //       p_ble_gattc_evt->gatt_status);
    return;
  }

  for (int i = 0; i < count; i++)
  {
    ble_usb_add_string("Characteristic Info: Char Handle = ", 35);
    ble_usb_add_value_uint16(p_ble_gattc_evt->params.char_disc_rsp.chars[i].handle_value, 1);
    ble_usb_add_string(", Char UUID = ", 14);
    ble_usb_add_value_uint16(p_ble_gattc_evt->params.char_disc_rsp.chars[i].uuid.uuid, 1);
    ble_usb_add_string(", Char Props = ", 15);
    ble_usb_add_char_props(p_ble_gattc_evt->params.char_disc_rsp.chars[i].char_props);
    ble_usb_add_string("\n", 1);
  }
}

/**@brief Function called on BLE_GATTC_EVT_DESC_DISC_RSP event.
 *
 * @details Update CCCD descriptor state and proceed to prompting user to toggle notifications.
 *
 * @param[in] p_ble_gattc_evt Descriptor Discovery Response Event.
 */
void on_descriptor_discovery_response(const ble_gattc_evt_t *const p_ble_gattc_evt)
{
  int count = p_ble_gattc_evt->params.desc_disc_rsp.count;

  // if (p_ble_gattc_evt->gatt_status == BLE_GATT_STATUS_ATTERR_ATTRIBUTE_NOT_FOUND)
  //{
  //     //printf("Descriptor discovery complete\n");
  //     return;
  // }
  // else if (p_ble_gattc_evt->gatt_status != NRF_SUCCESS)
  //{
  //     //printf("Descriptor discovery failed. Error code 0x%04X\n", p_ble_gattc_evt->gatt_status);
  //     return;
  // }

  // printf("Received descriptor discovery response, descriptor count: %d\n", count);

  for (int i = 0; i < 1; i++)
  {
    // printf("Descriptor handle: 0x%04X, UUID: 0x%04X\n",
    //        p_ble_gattc_evt->params.desc_disc_rsp.descs[i].handle,
    //        p_ble_gattc_evt->params.desc_disc_rsp.descs[i].uuid.uuid);

    // if (p_ble_gattc_evt->params.desc_disc_rsp.descs[i].uuid.uuid == BLE_UUID_CCCD)
    // {
    //   m_nus_cccd_handle = p_ble_gattc_evt->params.desc_disc_rsp.descs[i].handle;
    //   //printf("\'m_nus_cccd_handle\' assigned attribute handle 0x%02X\n", m_nus_cccd_handle);
    // }
    // else if (p_ble_gattc_evt->params.desc_disc_rsp.descs[i].uuid.uuid ==
    //          BLE_UUID_NUS_TX_CHARACTERISTIC)
    // {
    //   m_nus_tx_value_handle = p_ble_gattc_evt->params.desc_disc_rsp.descs[i].handle;
    //   //printf("\'m_nus_tx_value_handle\' assigned attribute handle 0x%02X\n",
    //          m_nus_tx_value_handle);
    // }
    // else if (p_ble_gattc_evt->params.desc_disc_rsp.descs[i].uuid.uuid ==
    //          BLE_UUID_NUS_RX_CHARACTERISTIC)
    // {
    //   m_nus_rx_value_handle = p_ble_gattc_evt->params.desc_disc_rsp.descs[i].handle;
    //   //printf("\'m_nus_rx_value_handle\' assigned attribute handle 0x%02X\n",
    //          m_nus_rx_value_handle);
    // }
    // Write data to serial port.
    ble_usb_add_string("Descriptor Handle = ", 21);
    ble_usb_add_value_uint16(p_ble_gattc_evt->params.desc_disc_rsp.descs[i].handle, 1);
    ble_usb_add_string(", Descriptor UUID = ", 20);
    ble_usb_add_value_uint16(p_ble_gattc_evt->params.desc_disc_rsp.descs->uuid.uuid, 1);
    ble_usb_add_string("\n", 1);
    ble_usb_printf();
  }
}

/**@brief Function called on BLE_GATTC_EVT_WRITE_RSP event.
 *
 * @param[in] p_ble_gattc_evt Write Response Event.
 */
void on_write_response(const ble_gattc_evt_t *const p_ble_gattc_evt)
{
  BLE_USB_PRINT_MSG("Received write response.\n");

  uint16_t len = p_ble_gattc_evt->params.write_rsp.len;
  ble_usb_add_string("Write response = 0x", 19);
  // for (uint16_t i = 0; i < len; i++)
  // {
  //   //printf("%02X", p_ble_gattc_evt->params.write_rsp.data[i]);
  // }
  ble_usb_add_array(p_ble_gattc_evt->params.write_rsp.data, p_ble_gattc_evt->params.write_rsp.len);
  ble_usb_add_string("\n", 1);
  ble_usb_printf();
  // Write data to serial port.

  if (p_ble_gattc_evt->gatt_status != NRF_SUCCESS)
  {
    // printf("Error. Write operation failed. Error code 0x%04X\n", p_ble_gattc_evt->gatt_status);
    BLE_USB_PRINT_MSG_NUM("Write Fail Error Code = 0x", p_ble_gattc_evt->gatt_status, 1);
  }
}

/**@brief Function called on BLE_GATTC_EVT_HVX event.
 *
 * @details Logs the received UART TX notifications.
 *
 * @param[in] p_ble_gattc_evt Handle Value Notification/Indication Event.
 */
void on_hvx(const ble_gattc_evt_t *const p_ble_gattc_evt)
{
  // ble_usb_add_string("HVX Handle = 0x", 15);
  uint16_t data_len = p_ble_gattc_evt->params.hvx.len;
  // ble_usb_add_value_uint16(data_len, 0);
  // ble_usb_add_string(". HVX Data = 0x", 13);
  // ble_usb_add_array(p_ble_gattc_evt->params.hvx.data, p_ble_gattc_evt->params.hvx.len);
  // ble_usb_add_string("\n", 1);
  // ble_usb_printf();
  usb_printf_variadic("HVX Handle = 0x%04X, HVX Data Length = %d, HVX Data = 0x", p_ble_gattc_evt->params.hvx.handle, data_len);
  for (size_t i = 0; i < data_len; i++)
  {
    // Write data to serial port.
    // printf("%02X ", p_ble_gattc_evt->params.hvx.data[i]);
    usb_printf_variadic("%02X ", p_ble_gattc_evt->params.hvx.data[i]);
  }
  // printf("\n");
  usb_printf_variadic("\n");
}

void on_read_response(ble_gattc_evt_t const *p_gattc_evt)
{
  // printf("Read 0x");
  ble_usb_add_string("Read Handle = ", 14);
  ble_usb_add_value_uint16(p_gattc_evt->params.read_rsp.handle, 1);
  ble_usb_add_string(", Read Data = 0x", 16);
  ble_usb_add_array(p_gattc_evt->params.read_rsp.data, p_gattc_evt->params.read_rsp.len);
  ble_usb_add_string("\n", 1);
  for (uint16_t i = 0; i < p_gattc_evt->params.read_rsp.len; i++)
  {
    // printf("%02X", p_gattc_evt->params.read_rsp.data[i]);
  }
  // printf("\n");
}

/**@brief Function called on BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST event.
 *
 * @details Update GAP connection parameters.
 *
 *
 * @param[in] p_ble_gap_evt Connection Parameter Update Event.
 */

void on_conn_params_update_request(const ble_gap_evt_t *const p_ble_gap_evt)
{
  uint32_t err_code = sd_ble_gap_conn_param_update(m_connection_handle,
                                                   &(p_ble_gap_evt->params.conn_param_update_request.conn_params));
  if (err_code != NRF_SUCCESS)
  {
    // printf("Conn params update failed, err_code %d\n", err_code);
    BLE_USB_PRINT_MSG_NUM("Conn Params Update Error Code = 0x", err_code, 1);
    return;
  }
  ble_usb_add_string("Conn Params = Min Conn Interval = ", 34);
  ble_usb_add_value_uint16(p_ble_gap_evt->params.conn_param_update_request.conn_params.min_conn_interval, 0);
  ble_usb_add_string(", Max Conn Interval = ", 22);
  ble_usb_add_value_uint16(p_ble_gap_evt->params.conn_param_update_request.conn_params.max_conn_interval, 0);
  ble_usb_add_string(", Slave Latency = ", 18);
  ble_usb_add_value_uint16(p_ble_gap_evt->params.conn_param_update_request.conn_params.slave_latency, 0);
  ble_usb_add_string(", Conn Superversion Timeout = ", 30);
  ble_usb_add_value_uint16(p_ble_gap_evt->params.conn_param_update_request.conn_params.conn_sup_timeout, 0);
  ble_usb_add_string("\n", 1);
  ble_usb_printf();
}

void on_char_val_by_uuid_read_rsp(ble_gattc_evt_t const *p_gattc_evt)
{
  uint16_t len = p_gattc_evt->params.read_rsp.len;
  // The above does not feel right, there should be no need to detect 00 as the end bit
  for (uint16_t i = 0; i < len; i++)
  {
    // printf("%02X", p_gattc_evt->params.read_rsp.data[i]);
  }
}

/**@brief Function called on BLE_GATTS_EVT_EXCHANGE_MTU_REQUEST event.
 *
 * @details Replies to an ATT_MTU exchange request by sending an Exchange MTU Response to the
 * client.
 *
 * @param[in] p_ble_gatts_evt Exchange MTU Request Event.
 */
void on_exchange_mtu_request(const ble_gatts_evt_t *const p_ble_gatts_evt)
{
  uint32_t err_code = sd_ble_gatts_exchange_mtu_reply(m_connection_handle,
                                                      NRF_SDH_BLE_GATT_MAX_MTU_SIZE);

  if (err_code != NRF_SUCCESS)
  {
    BLE_USB_PRINT_MSG_NUM("MTU Exchange Request Reply Error Code = 0x", err_code, 1);
  }
}

/**@brief Function called on BLE_GATTC_EVT_EXCHANGE_MTU_RSP event.
 *
 * @details Logs the new BLE server RX MTU size.
 *
 * @param[in] p_ble_gattc_evt Exchange MTU Response Event.
 */
void on_exchange_mtu_response(const ble_gattc_evt_t *const p_ble_gattc_evt)
{
  uint16_t server_rx_mtu = p_ble_gattc_evt->params.exchange_mtu_rsp.server_rx_mtu;

  BLE_USB_PRINT_MSG_NUM("MTU response received. New ATT_MTU is ", server_rx_mtu, 0);
}

void on_attr_info_disc(const ble_gattc_evt_t *const p_ble_gattc_evt)
{
  uint16_t count = p_ble_gattc_evt->params.attr_info_disc_rsp.count;
  if (count > 0)
  {
    // printf("Find %d attr_info\n", count);
    ble_usb_add_string("Find ", 5);
    ble_usb_add_value_uint16(count, 0);
    ble_usb_add_string(" Attr Info: ", 12);
    for (uint16_t i = 0; i < count; i++)
    {
      // printf("Attr format = %d\n", p_ble_gattc_evt->params.attr_info_disc_rsp.format);
      ble_usb_add_string("Attr", 4);
      ble_usb_add_value_uint16(i, 0);
      ble_usb_add_string("Format = ", 9);
      ble_usb_add_value_uint16(p_ble_gattc_evt->params.attr_info_disc_rsp.format, 0);
      ble_usb_add_string(", Handle = ", 11);
      if (p_ble_gattc_evt->params.attr_info_disc_rsp.format == 2)
      {
        // //printf("UUID: 0x");
        // for (int j = 0; j < 16; j++)
        // {
        //   //printf("%02X", p_ble_gattc_evt->params.attr_info_disc_rsp.info.attr_info128[i].uuid.uuid128[j]);
        // }
        // //printf("\n");
        ble_usb_add_value_uint16(p_ble_gattc_evt->params.attr_info_disc_rsp.info.attr_info128->handle, 1);
        ble_usb_add_string(", UUID = ", 9);
        ble_usb_add_array(p_ble_gattc_evt->params.attr_info_disc_rsp.info.attr_info128[i].uuid.uuid128, 16);
        ble_usb_add_string("\n", 1);
        ble_usb_printf();
      }
      else
      {
        // printf("Handle = 0x%02X, UUID = 0x%02X, Type = 0x%02X\n", p_ble_gattc_evt->params.attr_info_disc_rsp.info.attr_info16[i].handle, p_ble_gattc_evt->params.attr_info_disc_rsp.info.attr_info16[i].uuid.uuid, p_ble_gattc_evt->params.attr_info_disc_rsp.info.attr_info16[i].uuid.type);
        ble_usb_add_value_uint16(p_ble_gattc_evt->params.attr_info_disc_rsp.info.attr_info16->handle, 1);
        ble_usb_add_string(", UUID = ", 9);
        ble_usb_add_value_uint16(p_ble_gattc_evt->params.attr_info_disc_rsp.info.attr_info16[i].uuid.uuid, 1);
        ble_usb_add_string("\n", 1);
        ble_usb_printf();
      }
      //            _temp_attr_info128.handle =p_ble_gattc_evt->params.attr_info_disc_rsp.info.attr_info128->handle;
      //            _temp_attr_info128.uuid =p_ble_gattc_evt->params.attr_info_disc_rsp.info.attr_info128->uuid;
      // Write data to serial port.
    }
  }
}

void on_gatts_write(ble_gatts_evt_write_t write_evt)
{
  // printf("Gatts Handle 0x%02X Change: 0x", write_evt.handle);
  ble_usb_add_string("Gatts Write Handle = 0x", 23);
  ble_usb_add_value_uint16(write_evt.handle, 1);
  ble_usb_add_string(", Data = 0x", 11);
  ble_usb_add_array(write_evt.data, write_evt.len);
  ble_usb_add_string(", Offset = ", 11);
  ble_usb_add_value_uint16(write_evt.offset, 0);
  ble_usb_add_string("\n", 1);
  ble_usb_printf();
  // for (uint16_t i = 0; i < write_evt.len; i++)
  //  printf("%02X", write_evt.data[i]);
  //  printf(". Offset = %d\n", write_evt.offset);
}

void on_phy_requeset(const ble_gap_evt_t *const p_ble_gap_evt)
{
  usb_printf_variadic("PHY Update Request.\n");
  ble_gap_phys_t const phys = {BLE_GAP_PHY_AUTO, BLE_GAP_PHY_AUTO};
  uint32_t err_code = sd_ble_gap_phy_update(m_connection_handle, &phys);
  if (err_code != NRF_SUCCESS)
  {
    // printf("PHY update request reply failed, err_code %d\n", err_code);
    usb_printf_variadic("PHY Update Request Reply Error Code = 0x%04X\n", err_code);
  }
}

void ble_evt_dispatch(ble_evt_t *p_ble_evt, void *p_context)
{
  /* if (p_ble_evt == NULL)
  {
    printf("Received an empty BLE event\n");
    return;
  }
  // printf("Received BLE Event %d\n",p_ble_evt->header.evt_id);
  switch (p_ble_evt->header.evt_id)
  {
  case BLE_GAP_EVT_CONNECTED:
    on_connected(&(p_ble_evt->evt.gap_evt));
    break;

  case BLE_GAP_EVT_DISCONNECTED:
    printf("Disconnected, reason: 0x%02X\n",
           p_ble_evt->evt.gap_evt.params.disconnected.reason);
    BLE_USB_PRINT_MSG_NUM("Dissconnect Response: 0x", p_ble_evt->evt.gap_evt.params.disconnected.reason, 1);
    m_connected_devices--;
    m_connection_handle = BLE_CONN_HANDLE_INVALID;
    memset(m_connected_addr, 0, 6);
    // scan_start();
    break;

  case BLE_GAP_EVT_ADV_REPORT:
    on_adv_report(&(p_ble_evt->evt.gap_evt));
    break;

  case BLE_GAP_EVT_TIMEOUT:
    on_timeout(&(p_ble_evt->evt.gap_evt));
    break;

  case BLE_GATTC_EVT_PRIM_SRVC_DISC_RSP:
    on_service_discovery_response(&(p_ble_evt->evt.gattc_evt));
    break;

  case BLE_GATTC_EVT_CHAR_DISC_RSP:
    on_characteristic_discovery_response(&(p_ble_evt->evt.gattc_evt));
    break;

  case BLE_GATTC_EVT_DESC_DISC_RSP:
    on_descriptor_discovery_response(&(p_ble_evt->evt.gattc_evt));
    break;

  case BLE_GATTC_EVT_WRITE_RSP:
    on_write_response(&(p_ble_evt->evt.gattc_evt));
    break;

  case BLE_GATTC_EVT_HVX:
    on_hvx(&(p_ble_evt->evt.gattc_evt));
    break;

  case BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST:
    on_conn_params_update_request(&(p_ble_evt->evt.gap_evt));
    break;

  case BLE_GAP_EVT_CONN_PARAM_UPDATE:
  {
    ble_gap_conn_params_t *conn_params;

    conn_params = &p_ble_evt->evt.gap_evt.params.conn_param_update.conn_params;

    printf("Connection parameters updated. New parameters:\n");
    printf("Connection interval : %d [ms]\n",
           (int)(conn_params->min_conn_interval * 1.25));
    printf("Slave latency : %d\n", conn_params->slave_latency);
    printf("Supervision timeout : %d [s]\n",
           (int)(conn_params->conn_sup_timeout / 100));
    ble_usb_add_string("Conn Params = Min Conn Interval = ", 34);
    ble_usb_add_value_uint16(p_ble_evt->evt.gap_evt.params.conn_param_update_request.conn_params.min_conn_interval, 0);
    ble_usb_add_string(", Max Conn Interval = ", 22);
    ble_usb_add_value_uint16(p_ble_evt->evt.gap_evt.params.conn_param_update_request.conn_params.max_conn_interval, 0);
    ble_usb_add_string(", Slave Latency = ", 18);
    ble_usb_add_value_uint16(p_ble_evt->evt.gap_evt.params.conn_param_update_request.conn_params.slave_latency, 0);
    ble_usb_add_string(", Conn Superversion Timeout = ", 30);
    ble_usb_add_value_uint16(p_ble_evt->evt.gap_evt.params.conn_param_update_request.conn_params.conn_sup_timeout, 0);
    ble_usb_add_string("\n", 1);
    ble_usb_printf();
  }
  break;

  case BLE_GATTC_EVT_READ_RSP:
    on_read_response(&(p_ble_evt)->evt.gattc_evt);
    break;

  case BLE_GATTC_EVT_CHAR_VALS_READ_RSP:
    on_char_val_by_uuid_read_rsp(&(p_ble_evt)->evt.gattc_evt);
    break;
  case BLE_GATTC_EVT_CHAR_VAL_BY_UUID_READ_RSP:
    on_char_val_by_uuid_read_rsp(&(p_ble_evt)->evt.gattc_evt);
    break;

  case BLE_GATTC_EVT_ATTR_INFO_DISC_RSP:
    on_attr_info_disc(&(p_ble_evt)->evt.gattc_evt);
    break;

  case BLE_GAP_EVT_SEC_PARAMS_REQUEST:
    // on_sec_params_reply(p_ble_evt->evt.gap_evt.params.sec_params_request.peer_params);
    break;

    //        case BLE_GAP_EVT_AUTH_STATUS:
    //            ble_gap_evt_auth_status_t t_evt_status = p_ble_evt->evt.gap_evt.params.auth_status;
    //            if (t_evt_status.auth_status ==
    //                BLE_GAP_SEC_STATUS_SUCCESS)
    //            {
    //                printf("Pair Success\n");
    //            }
    //            else
    //            {
    //                uint8_t sm1_level =
    //                    (t_evt_status.sm1_levels.lv1 * 1) + (t_evt_status.sm1_levels.lv2 * 10) +
    //                    (t_evt_status.sm1_levels.lv3 * 100) + (t_evt_status.sm1_levels.lv4 * 1000);
    //                uint8_t sm2_level =
    //                    (t_evt_status.sm2_levels.lv1 * 1) + (t_evt_status.sm2_levels.lv2 * 10) +
    //                    (t_evt_status.sm2_levels.lv3 * 100) + (t_evt_status.sm2_levels.lv4 * 1000);
    //                printf("Error Pair, Status code = 0x%02X\nError Source = %d\nLESC Status = %d\nSM1 = %d\nSM2 = %d\n",
    //                            t_evt_status.auth_status,t_evt_status.error_src,t_evt_status.lesc,sm1_level,sm2_level);
    //            }
    //            break;

  case BLE_GAP_EVT_PASSKEY_DISPLAY:
    memcpy(passkey, p_ble_evt->evt.gap_evt.params.passkey_display.passkey, 6);
    break;

  case BLE_GAP_EVT_AUTH_KEY_REQUEST:
    // on_auth_key_request(p_ble_evt->evt.gap_evt);
    break;

  case BLE_GAP_EVT_LESC_DHKEY_REQUEST:
    // sd_ble_gap_lesc_dhkey_reply(m_adapter,m_connection_handle,&m_lesc_dhkey);
    // on_dhkey_reply(p_ble_evt->evt.gap_evt);
    break;

  case BLE_GAP_EVT_CONN_SEC_UPDATE:
    break;

  case BLE_GATTS_EVT_EXCHANGE_MTU_REQUEST:
    on_exchange_mtu_request(&(p_ble_evt->evt.gatts_evt));
    break;

  case BLE_GATTC_EVT_EXCHANGE_MTU_RSP:
    on_exchange_mtu_response(&(p_ble_evt->evt.gattc_evt));
    break;

  case BLE_GATTC_EVT_WRITE_CMD_TX_COMPLETE:
    break;

  case BLE_GAP_EVT_DATA_LENGTH_UPDATE:
    printf("Maximum radio packet length updated to %d bytes.\n",
           p_ble_evt->evt.gap_evt.params.data_length_update.effective_params
               .max_tx_octets);
    BLE_USB_PRINT_MSG_NUM("Maximum Radio Packet Length Updated To ", p_ble_evt->evt.gap_evt.params.data_length_update.effective_params.max_tx_octets, 0);
    break;

  case BLE_GAP_EVT_DATA_LENGTH_UPDATE_REQUEST:
    sd_ble_gap_data_length_update(m_connection_handle, NULL, NULL);
    break;

  case BLE_GAP_EVT_PHY_UPDATE_REQUEST:
  {
    printf("PHY update request.\n");
    BLE_USB_PRINT_MSG("PHY Update Request.\n");
    ble_gap_phys_t const phys = {BLE_GAP_PHY_AUTO, BLE_GAP_PHY_AUTO};
    uint32_t err_code = sd_ble_gap_phy_update(m_connection_handle, &phys);
    if (err_code != NRF_SUCCESS)
    {
      printf("PHY update request reply failed, err_code %d\n", err_code);
      BLE_USB_PRINT_MSG_NUM("PHY Update Request Reply Error Code = 0x", err_code, 1);
    }
  }
  break;
  case BLE_GATTS_EVT_SYS_ATTR_MISSING:
    sd_ble_gatts_sys_attr_set(m_connection_handle, NULL, 0, 0);
    break;
  case BLE_GATTS_EVT_WRITE:
    on_gatts_write(p_ble_evt->evt.gatts_evt.params.write);
    break;

  default:
    printf("Received an un-handled event with ID: 0x%04X\n",
           p_ble_evt->header.evt_id);
    break;
  }
  printf("Trigger Event: %d\n", p_ble_evt->header.evt_id);
  uint16_t event_len = BLE_EVT_BUF_SIZE; */
  BLE_USB_PRINT_MSG_NUM("Trigger Evt = ", p_ble_evt->header.evt_id, 0);
}